"""RNAFUSION upload API."""

import logging

import click

from cg.cli.upload.scout import scout
from cg.constants import DataDelivery
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.rnafusion import RnafusionAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Analysis, Family

LOG = logging.getLogger(__name__)


class RnafusionUploadAPI(UploadAPI):
    """RNAFUSION upload API."""

    def __init__(self, config: CGConfig):
        self.analysis_api: RnafusionAnalysisAPI = RnafusionAnalysisAPI(config)
        super().__init__(config=config, analysis_api=self.analysis_api)

    def upload(self, ctx: click.Context, case: Family, restart: bool) -> None:
        """Uploads RNAFUSION analysis data and files."""

        analysis: Analysis = case.analyses[0]
        self.update_upload_started_at(analysis)

        if DataDelivery.SCOUT in case.data_delivery:
            ctx.invoke(scout, case_id=case.internal_id, re_upload=restart)



"""Tests the findbusinessdata part of the Cg store API related to Analysis model."""
from datetime import datetime
from typing import List
from cg.store import Store
from cg.store.models import (
    Analysis,
)
from tests.store_helpers import StoreHelpers
from cg.constants import Pipeline


def test_get_analysis_by_case(
    store_with_case_and_analysis: Store, case_id: str = "test_case_internal_id"
):
    """Test that an analysis can be fetched by case."""
    # GIVEN a database with an analysis and case
    case = store_with_case_and_analysis.get_case_by_internal_id(internal_id=case_id)
    assert case
    # WHEN fetching the analysis by case
    analyses: List[Analysis] = store_with_case_and_analysis.get_analyses_by_case_entry_id(
        case_entry_id=case.id
    )
    # THEN one analysis should be returned
    for analysis in analyses:
        assert analysis
        assert analysis.family == case


def test_get_latest_nipt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store, timestamp_now: datetime
):
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN calling the analyses_to_delivery_report
    analyses = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_latest_nipt_analysis_to_upload()
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.family.internal_id in ["test_case_1", "yellowhog"]
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == Pipeline.FLUFFY


def test_get_latest_microsalt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_microsalt: Store, timestamp_now: datetime
):
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN calling the analyses_to_delivery_report
    analyses = (
        store_with_analyses_for_cases_not_uploaded_microsalt.get_latest_microsalt_analysis_to_upload()
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.family.internal_id in ["test_case_1", "yellowhog"]
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == Pipeline.MICROSALT



"""Tests for meta compress functionality that updates housekeeper"""

import logging
from pathlib import Path

from cg.constants import HK_FASTQ_TAGS
from cg.meta.compress import CompressAPI, files


def test_get_flow_cell_id_when_hiseqx(
    compress_api: CompressAPI, flow_cell_id: str, flow_cell_full_name: str
):
    """Test extracting the flow cell id from a fastq file path."""

    # GIVEN a CompressAPI and a flow cell id within a fastq file path
    fastq_path: Path = compress_api.demux_root.joinpath(
        Path(
            flow_cell_full_name, f"{flow_cell_id}-l6t11_Undetermined_GACGTCTT_L006_R1_001.fastq.gz"
        )
    )

    # WHEN retrieving the flow cell id
    returned_flow_cell_name: str = compress_api.get_flow_cell_id(fastq_path=fastq_path)

    # THEN the flow cell id retrieved should be identical to the flow cell id used
    assert returned_flow_cell_name == flow_cell_id


def test_get_flow_cell_id_when_novaseq(
    compress_api: CompressAPI, flow_cell_id: str, flow_cell_full_name: str
):
    """Test extracting the flow cell id from a fastq file path."""

    # GIVEN a CompressAPI and a flow cell id within a fastq file path
    fastq_path: Path = compress_api.demux_root.joinpath(
        Path(flow_cell_full_name, f"{flow_cell_id}_ACC10950A36_S36_L001_R1_001.fastq.gz")
    )

    # WHEN retrieving the flow cell id
    returned_flow_cell_name: str = compress_api.get_flow_cell_id(fastq_path=fastq_path)

    # THEN the flow cell id retrieved should be identical to the flow cell id used
    assert returned_flow_cell_name == flow_cell_id


def test_add_fastq_housekeeper_when_no_fastq_in_hk(
    caplog,
    compress_api,
    real_housekeeper_api,
    decompress_hk_spring_bundle,
    compression_files,
    store,
    helpers,
):
    """Test adding fastq files to Housekeeper when no fastq files in Housekeeper."""
    caplog.set_level(logging.INFO)

    # GIVEN real housekeeper api populated with a housekeeper bundle with spring info
    hk_bundle = decompress_hk_spring_bundle
    sample_id = hk_bundle["name"]
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    sample_obj = helpers.add_sample(store, internal_id=sample_id)
    compress_api.hk_api = real_housekeeper_api
    # GIVEN that there are no fastq files in HK
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    file_tags = set()
    for file_obj in version_obj.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)
    assert not set(HK_FASTQ_TAGS).intersection(file_tags)

    # WHEN adding the files to housekeeper
    compress_api.add_fastq_hk(
        sample_obj=sample_obj,
        fastq_first=compression_files.fastq_first_file,
        fastq_second=compression_files.fastq_second_file,
    )

    # THEN assert that the fastq files where added to HK
    version_obj = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    file_tags = set()
    for file_obj in version_obj.files:
        for tag in file_obj.tags:
            file_tags.add(tag.name)

    assert set(HK_FASTQ_TAGS).intersection(file_tags)


def test_add_decompressed_fastq(
    compress_api,
    real_housekeeper_api,
    decompress_hk_spring_bundle,
    compression_files,
    store,
    helpers,
):
    """Test functionality to add decompressed FASTQ files."""
    # GIVEN real Housekeeper api populated with a Housekeeper bundle with SPRING meta data
    hk_bundle = decompress_hk_spring_bundle
    sample_id = hk_bundle["name"]
    sample = helpers.add_sample(store, internal_id=sample_id)
    helpers.ensure_hk_bundle(real_housekeeper_api, hk_bundle)
    compress_api.hk_api = real_housekeeper_api

    # GIVEN that there exists a SPRING archive, spring metadata and unpacked fastqs
    version = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    fastq_first = compression_files.fastq_first_file
    fastq_second = compression_files.fastq_second_file
    spring_file = compression_files.spring_file
    spring_metadata_file = compression_files.updated_spring_metadata_file
    assert files.is_file_in_version(version_obj=version, path=spring_file)
    assert files.is_file_in_version(version_obj=version, path=spring_metadata_file)

    assert not files.is_file_in_version(version_obj=version, path=fastq_first)

    # WHEN adding decompressed files
    compress_api.add_decompressed_fastq(sample=sample)

    # THEN assert that the files where added
    version = compress_api.hk_api.get_latest_bundle_version(bundle_name=sample_id)
    assert files.is_file_in_version(version_obj=version, path=spring_file)
    assert files.is_file_in_version(version_obj=version, path=spring_metadata_file)

    assert files.is_file_in_version(version_obj=version, path=fastq_first)
    assert files.is_file_in_version(version_obj=version, path=fastq_second)



from typing import List, Tuple, Callable, Any
import pytest
from cg.utils.dispatcher import Dispatcher


def test_dispatch_table_generation(
    param1: str = None, param2: str = None, param3: str = None, param4: str = None
):
    def test_function_one(param1: str):
        pass

    def test_function_two(param1: str, param2: str):
        pass

    def test_function_three(param1: str, param2: str, param3: str):
        pass

    def test_function_four(param4: str):
        pass

    dispatcher = Dispatcher(
        [test_function_one, test_function_two, test_function_three, test_function_four],
        {"param1": param1, "param2": param2, "param3": param3, "param4": param4},
    )

    expected_table = {
        ("param1",): test_function_one,
        ("param1", "param2"): test_function_two,
        ("param1", "param2", "param3"): test_function_three,
        ("param4",): test_function_four,
    }
    assert dispatcher.dispatch_table.keys() == expected_table.keys()
    for key in expected_table:
        assert dispatcher.dispatch_table[key] == expected_table[key]


def test_call_matching_function(a: int = 1, b: int = 2, c: int = 3, x: int = 5, y: int = 4):
    def foo(a, b, c):
        return a + b + c

    def bar(x, y):
        return x * y

    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "x": x, "y": y})
    assert dispatcher({"a": a, "b": b, "c": c}) == 6
    assert dispatcher({"x": x, "y": y}) == 20


def test_call_non_matching_function(a: int = 1, b: int = 2, c: int = 3, x: int = 5, y: int = 4):
    def foo(a, b, c):
        return a + b + c

    def bar(x, y):
        return x * y

    dispatcher = Dispatcher([foo, bar], input_dict={"a": a, "b": b, "c": c, "x": x, "y": y})
    with pytest.raises(ValueError):
        dispatcher({"a": a, "b": b})


def test_call_for_function_without_parameters():
    def foo():
        return 1

    dispatcher = Dispatcher([foo], input_dict={})
    assert dispatcher({}) == 1


def test_call_for_function_with_same_number_of_parameters():
    def foo(a, b):
        return a + b

    def bar(c, d):
        return c * d

    dispatcher = Dispatcher([foo, bar], input_dict={"a": 1, "b": 2, "c": 3, "d": 4})
    assert dispatcher({"a": 1, "b": 2}) == 3
    assert dispatcher({"c": 3, "d": 4}) == 12


def test_call_dictionary_extra_parameters_not_in_functions():
    def foo(a, b):
        return a + b

    def bar(c, d):
        return c * d

    dispatcher = Dispatcher([foo, bar], input_dict={"a": 1, "b": 2, "c": 3, "d": 4, "e": 5})
    assert dispatcher({"a": 1, "b": 2}) == 3
